@@ -7,7 +7,7 @@
 
 #include <mgba/internal/gb/io.h>
 #include <mgba/internal/gb/mbc.h>
-#include <mgba/internal/lr35902/lr35902.h>
+#include <mgba/internal/sm83/sm83.h>
 
 #include <mgba/core/core.h>
 #include <mgba/core/cheats.h>
@@ -19,8 +19,8 @@
 
 #define CLEANUP_THRESHOLD 15
 
-const uint32_t CGB_LR35902_FREQUENCY = 0x800000;
-const uint32_t SGB_LR35902_FREQUENCY = 0x418B1E;
+const uint32_t CGB_SM83_FREQUENCY = 0x800000;
+const uint32_t SGB_SM83_FREQUENCY = 0x418B1E;
 
 const uint32_t GB_COMPONENT_MAGIC = 0x400000;
 
@@ -37,20 +37,15 @@ mLOG_DEFINE_CATEGORY(GB, "GB", "gb");
 
 static void GBInit(void* cpu, struct mCPUComponent* component);
 static void GBDeinit(struct mCPUComponent* component);
-static void GBInterruptHandlerInit(struct LR35902InterruptHandler* irqh);
-static void GBProcessEvents(struct LR35902Core* cpu);
-static void GBSetInterrupts(struct LR35902Core* cpu, bool enable);
-static uint16_t GBIRQVector(struct LR35902Core* cpu);
-static void GBIllegal(struct LR35902Core* cpu);
-static void GBStop(struct LR35902Core* cpu);
+static void GBInterruptHandlerInit(struct SM83InterruptHandler* irqh);
+static void GBProcessEvents(struct SM83Core* cpu);
+static void GBSetInterrupts(struct SM83Core* cpu, bool enable);
+static uint16_t GBIRQVector(struct SM83Core* cpu);
+static void GBIllegal(struct SM83Core* cpu);
+static void GBStop(struct SM83Core* cpu);
 
 static void _enableInterrupts(struct mTiming* timing, void* user, uint32_t cyclesLate);
 
-#ifdef FIXED_ROM_BUFFER
-extern uint32_t* romBuffer;
-extern size_t romBufferSize;
-#endif
-
 void GBCreate(struct GB* gb) {
 	gb->d.id = GB_COMPONENT_MAGIC;
 	gb->d.init = GBInit;
@@ -113,14 +108,7 @@ bool GBLoadROM(struct GB* gb, struct VFile* vf) {
 	gb->pristineRomSize = vf->size(vf);
 	vf->seek(vf, 0, SEEK_SET);
 	gb->isPristine = true;
-#ifdef FIXED_ROM_BUFFER
-	if (gb->pristineRomSize <= romBufferSize) {
-		gb->memory.rom = romBuffer;
-		vf->read(vf, romBuffer, gb->pristineRomSize);
-	}
-#else
 	gb->memory.rom = vf->map(vf, gb->pristineRomSize, MAP_READ);
-#endif
 	if (!gb->memory.rom) {
 		return false;
 	}
@@ -131,7 +119,7 @@ bool GBLoadROM(struct GB* gb, struct VFile* vf) {
 	GBMBCInit(gb);
 
 	if (gb->cpu) {
-		struct LR35902Core* cpu = gb->cpu;
+		struct SM83Core* cpu = gb->cpu;
 		cpu->memory.setActiveRegion(cpu, cpu->pc);
 	}
 
@@ -139,6 +127,19 @@ bool GBLoadROM(struct GB* gb, struct VFile* vf) {
 	return true;
 }
 
+void GBYankROM(struct GB* gb) {
+	gb->yankedRomSize = gb->memory.romSize;
+	gb->yankedMbc = gb->memory.mbcType;
+	gb->memory.romSize = 0;
+	gb->memory.mbcType = GB_MBC_NONE;
+	gb->memory.sramAccess = false;
+
+	if (gb->cpu) {
+		struct SM83Core* cpu = gb->cpu;
+		cpu->memory.setActiveRegion(cpu, cpu->pc);
+	}
+}
+
 static void GBSramDeinit(struct GB* gb) {
 	if (gb->sramVf) {
 		gb->sramVf->unmap(gb->sramVf, gb->memory.sram, gb->sramSize);
@@ -186,7 +187,7 @@ void GBResizeSram(struct GB* gb, size_t size) {
 					vf->write(vf, extdataBuffer, vfSize & 0xFF);
 				}
 				gb->memory.sram = vf->map(vf, size, MAP_WRITE);
-				memset(&gb->memory.sram[gb->sramSize], 0xFF, size - gb->sramSize);
+				memset(&gb->memory.sram[vfSize], 0xFF, size - vfSize);
 			} else if (size > gb->sramSize || !gb->memory.sram) {
 				if (gb->memory.sram) {
 					vf->unmap(vf, gb->memory.sram, gb->sramSize);
@@ -202,7 +203,7 @@ void GBResizeSram(struct GB* gb, size_t size) {
 		if (gb->memory.sram == (void*) -1) {
 			gb->memory.sram = NULL;
 		}
-	} else {
+	} else if (size) {
 		uint8_t* newSram = anonymousMemoryMap(size);
 		if (gb->memory.sram) {
 			if (size > gb->sramSize) {
@@ -374,7 +375,7 @@ void GBDestroy(struct GB* gb) {
 	mCoreCallbacksListDeinit(&gb->coreCallbacks);
 }
 
-void GBInterruptHandlerInit(struct LR35902InterruptHandler* irqh) {
+void GBInterruptHandlerInit(struct SM83InterruptHandler* irqh) {
 	irqh->reset = GBReset;
 	irqh->processEvents = GBProcessEvents;
 	irqh->setInterrupts = GBSetInterrupts;
@@ -409,35 +410,11 @@ bool GBIsBIOS(struct VFile* vf) {
 	}
 }
 
-void GBReset(struct LR35902Core* cpu) {
+void GBReset(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	gb->memory.romBase = gb->memory.rom;
 	GBDetectModel(gb);
 
-	if (gb->biosVf) {
-		if (!GBIsBIOS(gb->biosVf)) {
-			gb->biosVf->close(gb->biosVf);
-			gb->biosVf = NULL;
-		} else {
-			gb->biosVf->seek(gb->biosVf, 0, SEEK_SET);
-			gb->memory.romBase = malloc(GB_SIZE_CART_BANK0);
-			ssize_t size = gb->biosVf->read(gb->biosVf, gb->memory.romBase, GB_SIZE_CART_BANK0);
-			memcpy(&gb->memory.romBase[size], &gb->memory.rom[size], GB_SIZE_CART_BANK0 - size);
-			if (size > 0x100) {
-				memcpy(&gb->memory.romBase[0x100], &gb->memory.rom[0x100], sizeof(struct GBCartridge));
-			}
-
-			cpu->a = 0;
-			cpu->f.packed = 0;
-			cpu->c = 0;
-			cpu->e = 0;
-			cpu->h = 0;
-			cpu->l = 0;
-			cpu->sp = 0;
-			cpu->pc = 0;
-		}
-	}
-
 	cpu->b = 0;
 	cpu->d = 0;
 
@@ -449,6 +426,7 @@ void GBReset(struct LR35902Core* cpu) {
 
 	if (gb->yankedRomSize) {
 		gb->memory.romSize = gb->yankedRomSize;
+		gb->memory.mbcType = gb->yankedMbc;
 		gb->yankedRomSize = 0;
 	}
 
@@ -456,11 +434,30 @@ void GBReset(struct LR35902Core* cpu) {
 	gb->sgbControllers = 0;
 	gb->sgbCurrentController = 0;
 	gb->currentSgbBits = 0;
+	gb->sgbIncrement = false;
 	memset(gb->sgbPacket, 0, sizeof(gb->sgbPacket));
 
 	mTimingClear(&gb->timing);
 
 	GBMemoryReset(gb);
+
+	if (gb->biosVf) {
+		if (!GBIsBIOS(gb->biosVf)) {
+			gb->biosVf->close(gb->biosVf);
+			gb->biosVf = NULL;
+		} else {
+			GBMapBIOS(gb);
+			cpu->a = 0;
+			cpu->f.packed = 0;
+			cpu->c = 0;
+			cpu->e = 0;
+			cpu->h = 0;
+			cpu->l = 0;
+			cpu->sp = 0;
+			cpu->pc = 0;
+		}
+	}
+
 	GBVideoReset(&gb->video);
 	GBTimerReset(&gb->timer);
 	if (!gb->biosVf) {
@@ -480,7 +477,7 @@ void GBReset(struct LR35902Core* cpu) {
 }
 
 void GBSkipBIOS(struct GB* gb) {
-	struct LR35902Core* cpu = gb->cpu;
+	struct SM83Core* cpu = gb->cpu;
 	int nextDiv = 0;
 
 	switch (gb->model) {
@@ -563,10 +560,25 @@ void GBSkipBIOS(struct GB* gb) {
 	}
 }
 
+void GBMapBIOS(struct GB* gb) {
+	gb->biosVf->seek(gb->biosVf, 0, SEEK_SET);
+	uint8_t* oldRomBase = gb->memory.romBase;
+	gb->memory.romBase = malloc(GB_SIZE_CART_BANK0);
+	ssize_t size = gb->biosVf->read(gb->biosVf, gb->memory.romBase, GB_SIZE_CART_BANK0);
+	memcpy(&gb->memory.romBase[size], &oldRomBase[size], GB_SIZE_CART_BANK0 - size);
+	if (size > 0x100) {
+		memcpy(&gb->memory.romBase[0x100], &oldRomBase[0x100], sizeof(struct GBCartridge));
+	}
+}
+
 void GBUnmapBIOS(struct GB* gb) {
 	if (gb->memory.romBase < gb->memory.rom || gb->memory.romBase > &gb->memory.rom[gb->memory.romSize - 1]) {
 		free(gb->memory.romBase);
-		gb->memory.romBase = gb->memory.rom;
+		if (gb->memory.mbcType == GB_MMM01) {
+			GBMBCSwitchBank0(gb, gb->memory.romSize / GB_SIZE_CART_BANK0 - 2);
+		} else {
+			GBMBCSwitchBank0(gb, 0);
+		}
 	}
 	// XXX: Force AGB registers for AGB-mode
 	if (gb->model == GB_MODEL_AGB && gb->cpu->pc == 0x100) {
@@ -630,7 +642,7 @@ void GBDetectModel(struct GB* gb) {
 }
 
 void GBUpdateIRQs(struct GB* gb) {
-	int irqs = gb->memory.ie & gb->memory.io[REG_IF];
+	int irqs = gb->memory.ie & gb->memory.io[REG_IF] & 0x1F;
 	if (!irqs) {
 		gb->cpu->irqPending = false;
 		return;
@@ -644,10 +656,10 @@ void GBUpdateIRQs(struct GB* gb) {
 	if (gb->cpu->irqPending) {
 		return;
 	}
-	LR35902RaiseIRQ(gb->cpu);
+	SM83RaiseIRQ(gb->cpu);
 }
 
-void GBProcessEvents(struct LR35902Core* cpu) {
+void GBProcessEvents(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	do {
 		int32_t cycles = cpu->cycles;
@@ -675,7 +687,7 @@ void GBProcessEvents(struct LR35902Core* cpu) {
 	gb->earlyExit = false;
 }
 
-void GBSetInterrupts(struct LR35902Core* cpu, bool enable) {
+void GBSetInterrupts(struct SM83Core* cpu, bool enable) {
 	struct GB* gb = (struct GB*) cpu->master;
 	mTimingDeschedule(&gb->timing, &gb->eiPending);
 	if (!enable) {
@@ -686,7 +698,7 @@ void GBSetInterrupts(struct LR35902Core* cpu, bool enable) {
 	}
 }
 
-uint16_t GBIRQVector(struct LR35902Core* cpu) {
+uint16_t GBIRQVector(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	int irqs = gb->memory.ie & gb->memory.io[REG_IF];
 
@@ -721,9 +733,9 @@ static void _enableInterrupts(struct mTiming* timing, void* user, uint32_t cycle
 	GBUpdateIRQs(gb);
 }
 
-void GBHalt(struct LR35902Core* cpu) {
+void GBHalt(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
-	if (!(gb->memory.ie & gb->memory.io[REG_IF])) {
+	if (!(gb->memory.ie & gb->memory.io[REG_IF] & 0x1F)) {
 		cpu->cycles = cpu->nextEvent;
 		cpu->halted = true;
 	} else if (gb->model < GB_MODEL_CGB) {
@@ -731,7 +743,7 @@ void GBHalt(struct LR35902Core* cpu) {
 	}
 }
 
-void GBStop(struct LR35902Core* cpu) {
+void GBStop(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	if (cpu->bus) {
 		mLOG(GB, GAME_ERROR, "Hit illegal stop at address %04X:%02X", cpu->pc, cpu->bus);
@@ -746,7 +758,7 @@ void GBStop(struct LR35902Core* cpu) {
 		if (cpu->components && cpu->components[CPU_COMPONENT_DEBUGGER]) {
 			struct mDebuggerEntryInfo info = {
 				.address = cpu->pc - 1,
-				.type.bp.opcode = 0x1000 | cpu->bus
+				.type.bp.opcode = 0x1000 | cpu->bus,
 			};
 			mDebuggerEnter((struct mDebugger*) cpu->components[CPU_COMPONENT_DEBUGGER], DEBUGGER_ENTER_ILLEGAL_OP, &info);
 		}
@@ -758,7 +770,7 @@ void GBStop(struct LR35902Core* cpu) {
 	// TODO: Actually stop
 }
 
-void GBIllegal(struct LR35902Core* cpu) {
+void GBIllegal(struct SM83Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	mLOG(GB, GAME_ERROR, "Hit illegal opcode at address %04X:%02X", cpu->pc, cpu->bus);
 #ifdef USE_DEBUGGERS
