@@ -6,7 +6,7 @@
 #include <mgba/internal/gb/mbc.h>
 
 #include <mgba/core/interface.h>
-#include <mgba/internal/lr35902/lr35902.h>
+#include <mgba/internal/sm83/sm83.h>
 #include <mgba/internal/gb/gb.h>
 #include <mgba/internal/gb/memory.h>
 #include <mgba-util/crc32.h>
@@ -35,6 +35,7 @@ static void _GBHuC1(struct GB*, uint16_t address, uint8_t value);
 static void _GBHuC3(struct GB*, uint16_t address, uint8_t value);
 static void _GBPocketCam(struct GB* gb, uint16_t address, uint8_t value);
 static void _GBTAMA5(struct GB* gb, uint16_t address, uint8_t value);
+static void _GBWisdomTree(struct GB* gb, uint16_t address, uint8_t value);
 
 static uint8_t _GBMBC2Read(struct GBMemory*, uint16_t address);
 static uint8_t _GBMBC6Read(struct GBMemory*, uint16_t address);
@@ -119,6 +120,29 @@ static bool _isMulticart(const uint8_t* mem) {
 	return success;
 }
 
+static bool _isWisdomTree(const uint8_t* mem, size_t size) {
+	size_t i;
+	for (i = 0x134; i < 0x14C; i += 4) {
+		if (*(uint32_t*) &mem[i] != 0) {
+			return false;
+		}
+	}
+	for (i = 0xF0; i < 0x100; i += 4) {
+		if (*(uint32_t*) &mem[i] != 0) {
+			return false;
+		}
+	}
+	if (mem[0x14D] != 0xE7) {
+		return false;
+	}
+	for (i = 0x300; i < size - 11; ++i) {
+		if (memcmp(&mem[i], "WISDOM", 6) == 0 && memcmp(&mem[i + 7], "TREE", 4) == 0) {
+			return true;
+		}
+	}
+	return false;
+}
+
 void GBMBCSwitchSramBank(struct GB* gb, int bank) {
 	size_t bankStart = bank * GB_SIZE_EXTERNAL_RAM;
 	if (bankStart + GB_SIZE_EXTERNAL_RAM > gb->sramSize) {
@@ -180,6 +204,11 @@ void GBMBCInit(struct GB* gb) {
 		if (gb->memory.mbcType == GB_MBC_AUTODETECT) {
 			switch (cart->type) {
 			case 0:
+				if (_isWisdomTree(gb->memory.rom, gb->memory.romSize)) {
+					gb->memory.mbcType = GB_UNL_WISDOM_TREE;
+					break;
+				}
+				// Fall through
 			case 8:
 			case 9:
 				gb->memory.mbcType = GB_MBC_NONE;
@@ -310,6 +339,9 @@ void GBMBCInit(struct GB* gb) {
 			gb->memory.cam->startRequestImage(gb->memory.cam, GBCAM_WIDTH, GBCAM_HEIGHT, mCOLOR_ANY);
 		}
 		break;
+	case GB_UNL_WISDOM_TREE:
+		gb->memory.mbcWrite = _GBWisdomTree;
+		break;
 	}
 
 	gb->memory.currentBank = 1;
@@ -1167,6 +1199,21 @@ uint8_t _GBTAMA5Read(struct GBMemory* memory, uint16_t address) {
 	}
 }
 
+void _GBWisdomTree(struct GB* gb, uint16_t address, uint8_t value) {
+	UNUSED(value);
+	int bank = address & 0x3F;
+	switch (address >> 14) {
+	case 0x0:
+		GBMBCSwitchBank0(gb, bank * 2);
+		GBMBCSwitchBank(gb, bank * 2 + 1);
+		break;
+	default:
+		// TODO
+		mLOG(GB_MBC, STUB, "Wisdom Tree unknown address: %04X:%02X", address, value);
+		break;
+	}
+}
+
 void GBMBCRTCRead(struct GB* gb) {
 	struct GBMBCRTCSaveBuffer rtcBuffer;
 	struct VFile* vf = gb->sramVf;
