@@ -12,7 +12,7 @@
 #include <mgba/internal/gb/io.h>
 #include <mgba/internal/gb/renderers/cache-set.h>
 #include <mgba/internal/gb/serialize.h>
-#include <mgba/internal/lr35902/lr35902.h>
+#include <mgba/internal/sm83/sm83.h>
 
 #include <mgba-util/memory.h>
 
@@ -187,7 +187,7 @@ void GBVideoAssociateRenderer(struct GBVideo* video, struct GBVideoRenderer* ren
 	video->renderer->init(video->renderer, video->p->model, video->sgbBorders);
 }
 
-static bool _statIRQAsserted(struct GBVideo* video, GBRegisterSTAT stat) {
+static bool _statIRQAsserted(GBRegisterSTAT stat) {
 	// TODO: variable for the IRQ line value?
 	if (GBRegisterSTATIsLYCIRQ(stat) && GBRegisterSTATIsLYC(stat)) {
 		return true;
@@ -224,7 +224,6 @@ void _endMode0(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	++video->ly;
 	video->p->memory.io[REG_LY] = video->ly;
 	GBRegisterSTAT oldStat = video->stat;
-	video->stat = GBRegisterSTATSetLYC(video->stat, lyc == video->ly);
 	if (video->ly < GB_VIDEO_VERTICAL_PIXELS) {
 		next = GB_VIDEO_MODE_2_LENGTH;
 		video->mode = 2;
@@ -237,15 +236,23 @@ void _endMode0(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 		mTimingDeschedule(&video->p->timing, &video->frameEvent);
 		mTimingSchedule(&video->p->timing, &video->frameEvent, -cyclesLate);
 
-		if (!_statIRQAsserted(video, oldStat) && GBRegisterSTATIsOAMIRQ(video->stat)) {
+		if (!_statIRQAsserted(oldStat) && GBRegisterSTATIsOAMIRQ(video->stat)) {
 			video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 		}
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_VBLANK);
 	}
 	video->stat = GBRegisterSTATSetMode(video->stat, video->mode);
-	if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+	if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 	}
+
+	// LYC stat is delayed 1 T-cycle
+	oldStat = video->stat;
+	video->stat = GBRegisterSTATSetLYC(video->stat, lyc == video->ly);
+	if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
+		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
+	}
+
 	GBUpdateIRQs(video->p);
 	video->p->memory.io[REG_STAT] = video->stat;
 	mTimingSchedule(timing, &video->modeEvent, (next << video->p->doubleSpeed) - cyclesLate);
@@ -280,7 +287,7 @@ void _endMode1(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	GBRegisterSTAT oldStat = video->stat;
 	video->stat = GBRegisterSTATSetMode(video->stat, video->mode);
 	video->stat = GBRegisterSTATSetLYC(video->stat, lyc == video->p->memory.io[REG_LY]);
-	if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+	if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 		GBUpdateIRQs(video->p);
 	}
@@ -298,7 +305,7 @@ void _endMode2(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	video->modeEvent.callback = _endMode3;
 	GBRegisterSTAT oldStat = video->stat;
 	video->stat = GBRegisterSTATSetMode(video->stat, video->mode);
-	if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+	if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 		GBUpdateIRQs(video->p);
 	}
@@ -319,7 +326,7 @@ void _endMode3(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	video->modeEvent.callback = _endMode0;
 	GBRegisterSTAT oldStat = video->stat;
 	video->stat = GBRegisterSTATSetMode(video->stat, video->mode);
-	if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+	if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 		GBUpdateIRQs(video->p);
 	}
@@ -332,23 +339,23 @@ void _endMode3(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 void _updateFrameCount(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	UNUSED(cyclesLate);
 	struct GBVideo* video = context;
-	if (video->p->cpu->executionState != LR35902_CORE_FETCH) {
+	if (video->p->cpu->executionState != SM83_CORE_FETCH) {
 		mTimingSchedule(timing, &video->frameEvent, 4 - ((video->p->cpu->executionState + 1) & 3));
 		return;
 	}
+	if (!GBRegisterLCDCIsEnable(video->p->memory.io[REG_LCDC])) {
+		mTimingSchedule(timing, &video->frameEvent, GB_VIDEO_TOTAL_LENGTH);
+	}
 
-	GBFrameEnded(video->p);
-	mCoreSyncPostFrame(video->p->sync);
 	--video->frameskipCounter;
 	if (video->frameskipCounter < 0) {
 		video->renderer->finishFrame(video->renderer);
 		video->frameskipCounter = video->frameskip;
 	}
+	GBFrameEnded(video->p);
+	mCoreSyncPostFrame(video->p->sync);
 	++video->frameCounter;
 
-	if (!GBRegisterLCDCIsEnable(video->p->memory.io[REG_LCDC])) {
-		mTimingSchedule(timing, &video->frameEvent, GB_VIDEO_TOTAL_LENGTH);
-	}
 	GBFrameStarted(video->p);
 }
 
@@ -409,7 +416,7 @@ void GBVideoWriteLCDC(struct GBVideo* video, GBRegisterLCDC value) {
 		GBRegisterSTAT oldStat = video->stat;
 		video->stat = GBRegisterSTATSetMode(video->stat, 0);
 		video->stat = GBRegisterSTATSetLYC(video->stat, video->ly == video->p->memory.io[REG_LYC]);
-		if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+		if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 			video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 			GBUpdateIRQs(video->p);
 		}
@@ -440,7 +447,7 @@ void GBVideoWriteSTAT(struct GBVideo* video, GBRegisterSTAT value) {
 	if (!GBRegisterLCDCIsEnable(video->p->memory.io[REG_LCDC]) || video->p->model >= GB_MODEL_CGB) {
 		return;
 	}
-	if (!_statIRQAsserted(video, oldStat) && video->mode < 3) {
+	if (!_statIRQAsserted(oldStat) && video->mode < 3) {
 		// TODO: variable for the IRQ line value?
 		video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 		GBUpdateIRQs(video->p);
@@ -451,7 +458,7 @@ void GBVideoWriteLYC(struct GBVideo* video, uint8_t value) {
 	GBRegisterSTAT oldStat = video->stat;
 	if (GBRegisterLCDCIsEnable(video->p->memory.io[REG_LCDC])) {
 		video->stat = GBRegisterSTATSetLYC(video->stat, value == video->ly);
-		if (!_statIRQAsserted(video, oldStat) && _statIRQAsserted(video, video->stat)) {
+		if (!_statIRQAsserted(oldStat) && _statIRQAsserted(video->stat)) {
 			video->p->memory.io[REG_IF] |= (1 << GB_IRQ_LCDSTAT);
 			GBUpdateIRQs(video->p);
 		}
@@ -498,14 +505,16 @@ void GBVideoWritePalette(struct GBVideo* video, uint16_t address, uint8_t value)
 	} else {
 		switch (address) {
 		case REG_BCPD:
-			if (video->bcpIndex & 1) {
-				video->palette[video->bcpIndex >> 1] &= 0x00FF;
-				video->palette[video->bcpIndex >> 1] |= value << 8;
-			} else {
-				video->palette[video->bcpIndex >> 1] &= 0xFF00;
-				video->palette[video->bcpIndex >> 1] |= value;
+			if (video->mode != 3) {
+				if (video->bcpIndex & 1) {
+					video->palette[video->bcpIndex >> 1] &= 0x00FF;
+					video->palette[video->bcpIndex >> 1] |= value << 8;
+				} else {
+					video->palette[video->bcpIndex >> 1] &= 0xFF00;
+					video->palette[video->bcpIndex >> 1] |= value;
+				}
+				video->renderer->writePalette(video->renderer, video->bcpIndex >> 1, video->palette[video->bcpIndex >> 1]);
 			}
-			video->renderer->writePalette(video->renderer, video->bcpIndex >> 1, video->palette[video->bcpIndex >> 1]);
 			if (video->bcpIncrement) {
 				++video->bcpIndex;
 				video->bcpIndex &= 0x3F;
@@ -515,14 +524,16 @@ void GBVideoWritePalette(struct GBVideo* video, uint16_t address, uint8_t value)
 			video->p->memory.io[REG_BCPD] = video->palette[video->bcpIndex >> 1] >> (8 * (video->bcpIndex & 1));
 			break;
 		case REG_OCPD:
-			if (video->ocpIndex & 1) {
-				video->palette[8 * 4 + (video->ocpIndex >> 1)] &= 0x00FF;
-				video->palette[8 * 4 + (video->ocpIndex >> 1)] |= value << 8;
-			} else {
-				video->palette[8 * 4 + (video->ocpIndex >> 1)] &= 0xFF00;
-				video->palette[8 * 4 + (video->ocpIndex >> 1)] |= value;
+			if (video->mode != 3) {
+				if (video->ocpIndex & 1) {
+					video->palette[8 * 4 + (video->ocpIndex >> 1)] &= 0x00FF;
+					video->palette[8 * 4 + (video->ocpIndex >> 1)] |= value << 8;
+				} else {
+					video->palette[8 * 4 + (video->ocpIndex >> 1)] &= 0xFF00;
+					video->palette[8 * 4 + (video->ocpIndex >> 1)] |= value;
+				}
+				video->renderer->writePalette(video->renderer, 8 * 4 + (video->ocpIndex >> 1), video->palette[8 * 4 + (video->ocpIndex >> 1)]);
 			}
-			video->renderer->writePalette(video->renderer, 8 * 4 + (video->ocpIndex >> 1), video->palette[8 * 4 + (video->ocpIndex >> 1)]);
 			if (video->ocpIncrement) {
 				++video->ocpIndex;
 				video->ocpIndex &= 0x3F;
@@ -690,8 +701,11 @@ void GBVideoWriteSGBPacket(struct GBVideo* video, uint8_t* data) {
 	case SGB_ATTR_SET:
 		break;
 	case SGB_MLT_REQ:
+		if ((video->sgbPacketBuffer[1] & 0x3) == 2) { // XXX: This unmasked increment appears to be an SGB hardware bug
+			++video->p->sgbCurrentController;
+		}
 		video->p->sgbControllers = video->sgbPacketBuffer[1] & 0x3;
-		video->p->sgbCurrentController = 0;
+		video->p->sgbCurrentController &= video->p->sgbControllers;
 		return;
 	case SGB_MASK_EN:
 		video->renderer->sgbRenderMode = video->sgbPacketBuffer[1] & 0x3;
