@@ -461,19 +461,19 @@ void GBAIOWrite(struct GBA* gba, uint32_t address, uint16_t value) {
 			break;
 
 		case REG_DMA0CNT_LO:
-			GBADMAWriteCNT_LO(gba, 0, value);
+			GBADMAWriteCNT_LO(gba, 0, value & 0x3FFF);
 			break;
 		case REG_DMA0CNT_HI:
 			value = GBADMAWriteCNT_HI(gba, 0, value);
 			break;
 		case REG_DMA1CNT_LO:
-			GBADMAWriteCNT_LO(gba, 1, value);
+			GBADMAWriteCNT_LO(gba, 1, value & 0x3FFF);
 			break;
 		case REG_DMA1CNT_HI:
 			value = GBADMAWriteCNT_HI(gba, 1, value);
 			break;
 		case REG_DMA2CNT_LO:
-			GBADMAWriteCNT_LO(gba, 2, value);
+			GBADMAWriteCNT_LO(gba, 2, value & 0x3FFF);
 			break;
 		case REG_DMA2CNT_HI:
 			value = GBADMAWriteCNT_HI(gba, 2, value);
@@ -547,15 +547,18 @@ void GBAIOWrite(struct GBA* gba, uint32_t address, uint16_t value) {
 			GBAAdjustWaitstates(gba, value);
 			break;
 		case REG_IE:
-			GBAWriteIE(gba, value);
-			break;
+			gba->memory.io[REG_IE >> 1] = value;
+			GBATestIRQ(gba, 1);
+			return;
 		case REG_IF:
-			gba->springIRQ &= ~value;
 			value = gba->memory.io[REG_IF >> 1] & ~value;
-			break;
+			gba->memory.io[REG_IF >> 1] = value;
+			GBATestIRQ(gba, 1);
+			return;
 		case REG_IME:
-			GBAWriteIME(gba, value);
-			break;
+			gba->memory.io[REG_IME >> 1] = value;
+			GBATestIRQ(gba, 1);
+			return;
 		case REG_MAX:
 			// Some bad interrupt libraries will write to this
 			break;
@@ -709,19 +712,28 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 	switch (address) {
 	// Reading this takes two cycles (1N+1I), so let's remove them preemptively
 	case REG_TM0CNT_LO:
-		GBATimerUpdateRegister(gba, 0, 4);
+		GBATimerUpdateRegister(gba, 0, 2);
 		break;
 	case REG_TM1CNT_LO:
-		GBATimerUpdateRegister(gba, 1, 4);
+		GBATimerUpdateRegister(gba, 1, 2);
 		break;
 	case REG_TM2CNT_LO:
-		GBATimerUpdateRegister(gba, 2, 4);
+		GBATimerUpdateRegister(gba, 2, 2);
 		break;
 	case REG_TM3CNT_LO:
-		GBATimerUpdateRegister(gba, 3, 4);
+		GBATimerUpdateRegister(gba, 3, 2);
 		break;
 
 	case REG_KEYINPUT:
+		{
+			size_t c;
+			for (c = 0; c < mCoreCallbacksListSize(&gba->coreCallbacks); ++c) {
+				struct mCoreCallbacks* callbacks = mCoreCallbacksListGetPointer(&gba->coreCallbacks, c);
+				if (callbacks->keysRead) {
+					callbacks->keysRead(callbacks->context);
+				}
+			}
+		}
 		if (gba->rr && gba->rr->isPlaying(gba->rr)) {
 			return 0x3FF ^ gba->rr->queryInput(gba->rr);
 		} else {
@@ -814,8 +826,10 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 	case REG_DMA1CNT_LO:
 	case REG_DMA2CNT_LO:
 	case REG_DMA3CNT_LO:
-		// Write-only register
-		mLOG(GBA_IO, GAME_ERROR, "Read from write-only I/O register: %03X", address);
+		// Many, many things read from the DMA register
+	case REG_MAX:
+		// Some bad interrupt libraries will read from this
+		// (Silent) write-only register
 		return 0;
 
 	case REG_JOY_RECV_LO:
@@ -888,8 +902,6 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 	case REG_IME:
 		// Handled transparently by registers
 		break;
-	case REG_MAX:
-		// Some bad interrupt libraries will read from this
 	case 0x066:
 	case 0x06E:
 	case 0x076:
@@ -931,7 +943,6 @@ void GBAIOSerialize(struct GBA* gba, struct GBASerializedState* state) {
 		STORE_16(gba->timers[i].reload, 0, &state->timers[i].reload);
 		STORE_32(gba->timers[i].lastEvent - mTimingCurrentTime(&gba->timing), 0, &state->timers[i].lastEvent);
 		STORE_32(gba->timers[i].event.when - mTimingCurrentTime(&gba->timing), 0, &state->timers[i].nextEvent);
-		STORE_32(gba->timers[i].irq.when - mTimingCurrentTime(&gba->timing), 0, &state->timers[i].nextIrq);
 		STORE_32(gba->timers[i].flags, 0, &state->timers[i].flags);
 		STORE_32(gba->memory.dma[i].nextSource, 0, &state->dma[i].nextSource);
 		STORE_32(gba->memory.dma[i].nextDest, 0, &state->dma[i].nextDest);
@@ -939,7 +950,7 @@ void GBAIOSerialize(struct GBA* gba, struct GBASerializedState* state) {
 		STORE_32(gba->memory.dma[i].when, 0, &state->dma[i].when);
 	}
 
-	state->dmaTransferRegister = gba->memory.dmaTransferRegister;
+	STORE_32(gba->memory.dmaTransferRegister, 0, &state->dmaTransferRegister);
 
 	GBAHardwareSerialize(&gba->memory.hw, state);
 }
@@ -971,22 +982,17 @@ void GBAIODeserialize(struct GBA* gba, const struct GBASerializedState* state) {
 		if (GBATimerFlagsIsEnable(gba->timers[i].flags)) {
 			mTimingSchedule(&gba->timing, &gba->timers[i].event, when);
 		}
-		LOAD_32(when, 0, &state->timers[i].nextIrq);
-		if (GBATimerFlagsIsIrqPending(gba->timers[i].flags)) {
-			mTimingSchedule(&gba->timing, &gba->timers[i].irq, when);
-		}
 
 		LOAD_16(gba->memory.dma[i].reg, (REG_DMA0CNT_HI + i * 12), state->io);
 		LOAD_32(gba->memory.dma[i].nextSource, 0, &state->dma[i].nextSource);
 		LOAD_32(gba->memory.dma[i].nextDest, 0, &state->dma[i].nextDest);
 		LOAD_32(gba->memory.dma[i].nextCount, 0, &state->dma[i].nextCount);
 		LOAD_32(gba->memory.dma[i].when, 0, &state->dma[i].when);
-		if (GBADMARegisterGetTiming(gba->memory.dma[i].reg) != GBA_DMA_TIMING_NOW) {
-			GBADMASchedule(gba, i, &gba->memory.dma[i]);
-		}
 	}
 	GBAAudioWriteSOUNDCNT_X(&gba->audio, gba->memory.io[REG_SOUNDCNT_X >> 1]);
-	gba->memory.dmaTransferRegister = state->dmaTransferRegister;
+
+	LOAD_32(gba->memory.dmaTransferRegister, 0, &state->dmaTransferRegister);
+
 	GBADMAUpdate(gba);
 	GBAHardwareDeserialize(&gba->memory.hw, state);
 }
