@@ -77,6 +77,9 @@ void GBASavedataDeinit(struct GBASavedata* savedata) {
 		case SAVEDATA_EEPROM:
 			mappedMemoryFree(savedata->data, SIZE_CART_EEPROM);
 			break;
+		case SAVEDATA_EEPROM512:
+			mappedMemoryFree(savedata->data, SIZE_CART_EEPROM512);
+			break;
 		case SAVEDATA_FORCE_NONE:
 		case SAVEDATA_AUTODETECT:
 			break;
@@ -127,6 +130,8 @@ bool GBASavedataClone(struct GBASavedata* savedata, struct VFile* out) {
 			return out->write(out, savedata->data, SIZE_CART_FLASH1M) == SIZE_CART_FLASH1M;
 		case SAVEDATA_EEPROM:
 			return out->write(out, savedata->data, SIZE_CART_EEPROM) == SIZE_CART_EEPROM;
+		case SAVEDATA_EEPROM512:
+			return out->write(out, savedata->data, SIZE_CART_EEPROM512) == SIZE_CART_EEPROM512;
 		case SAVEDATA_AUTODETECT:
 		case SAVEDATA_FORCE_NONE:
 			return true;
@@ -152,7 +157,9 @@ size_t GBASavedataSize(const struct GBASavedata* savedata) {
 	case SAVEDATA_FLASH1M:
 		return SIZE_CART_FLASH1M;
 	case SAVEDATA_EEPROM:
-		return (savedata->vf && savedata->vf->size(savedata->vf) == SIZE_CART_EEPROM512) ? SIZE_CART_EEPROM512 : SIZE_CART_EEPROM;
+		return SIZE_CART_EEPROM;
+	case SAVEDATA_EEPROM512:
+		return SIZE_CART_EEPROM512;
 	case SAVEDATA_FORCE_NONE:
 		return 0;
 	case SAVEDATA_AUTODETECT:
@@ -218,6 +225,8 @@ void GBASavedataForceType(struct GBASavedata* savedata, enum SavedataType type)
 		GBASavedataInitFlash(savedata);
 		break;
 	case SAVEDATA_EEPROM:
+	case SAVEDATA_EEPROM512:
+		savedata->type = type;
 		GBASavedataInitEEPROM(savedata);
 		break;
 	case SAVEDATA_SRAM:
@@ -263,22 +272,23 @@ void GBASavedataInitFlash(struct GBASavedata* savedata) {
 
 void GBASavedataInitEEPROM(struct GBASavedata* savedata) {
 	if (savedata->type == SAVEDATA_AUTODETECT) {
-		savedata->type = SAVEDATA_EEPROM;
-	} else {
+		savedata->type = SAVEDATA_EEPROM512;
+	} else if (savedata->type != SAVEDATA_EEPROM512 && savedata->type != SAVEDATA_EEPROM) {
 		mLOG(GBA_SAVE, WARN, "Can't re-initialize savedata");
 		return;
 	}
 	int32_t eepromSize = SIZE_CART_EEPROM512;
+	if (savedata->type == SAVEDATA_EEPROM) {
+		eepromSize = SIZE_CART_EEPROM;
+	}
 	off_t end;
 	if (!savedata->vf) {
 		end = 0;
 		savedata->data = anonymousMemoryMap(SIZE_CART_EEPROM);
 	} else {
 		end = savedata->vf->size(savedata->vf);
-		if (end < SIZE_CART_EEPROM512) {
-			savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM512);
-		} else if (end > SIZE_CART_EEPROM512) {
-			eepromSize = SIZE_CART_EEPROM;
+		if (end < eepromSize) {
+			savedata->vf->truncate(savedata->vf, eepromSize);
 		}
 		savedata->data = savedata->vf->map(savedata->vf, eepromSize, savedata->mapMode);
 	}
@@ -420,13 +430,21 @@ static void _ensureEeprom(struct GBASavedata* savedata, uint32_t size) {
 	if (size < SIZE_CART_EEPROM512) {
 		return;
 	}
-	if (!savedata->vf || savedata->vf->size(savedata->vf) > SIZE_CART_EEPROM512) {
+	if (savedata->type == SAVEDATA_EEPROM) {
+		return;
+	}
+	savedata->type = SAVEDATA_EEPROM;
+	if (!savedata->vf) {
 		return;
 	}
 	savedata->vf->unmap(savedata->vf, savedata->data, SIZE_CART_EEPROM512);
-	savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM);
-	savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_EEPROM, savedata->mapMode);
-	memset(&savedata->data[SIZE_CART_EEPROM512], 0xFF, SIZE_CART_EEPROM - SIZE_CART_EEPROM512);
+	if (savedata->vf->size(savedata->vf) < SIZE_CART_EEPROM) {
+		savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM);
+		savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_EEPROM, savedata->mapMode);
+		memset(&savedata->data[SIZE_CART_EEPROM512], 0xFF, SIZE_CART_EEPROM - SIZE_CART_EEPROM512);
+	} else {
+		savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_EEPROM, savedata->mapMode);
+	}
 }
 
 void GBASavedataWriteEEPROM(struct GBASavedata* savedata, uint16_t value, uint32_t writeSize) {
@@ -586,7 +604,7 @@ void _flashSwitchBank(struct GBASavedata* savedata, int bank) {
 		savedata->type = SAVEDATA_FLASH1M;
 		if (savedata->vf) {
 			savedata->vf->unmap(savedata->vf, savedata->data, SIZE_CART_FLASH512);
-			if (savedata->vf->size(savedata->vf) == SIZE_CART_FLASH512) {
+			if (savedata->vf->size(savedata->vf) < SIZE_CART_FLASH1M) {
 				savedata->vf->truncate(savedata->vf, SIZE_CART_FLASH1M);
 				savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_FLASH1M, MAP_WRITE);
 				memset(&savedata->data[SIZE_CART_FLASH512], 0xFF, SIZE_CART_FLASH512);
