@@ -40,16 +40,6 @@ CoreController::CoreController(mCore* core, QObject* parent)
 	m_threadContext.core = core;
 	m_threadContext.userData = this;
 
-	QSize size(256, 512);
-	m_buffers[0].resize(size.width() * size.height() * sizeof(color_t));
-	m_buffers[1].resize(size.width() * size.height() * sizeof(color_t));
-	m_buffers[0].fill(0xFF);
-	m_buffers[1].fill(0xFF);
-	m_activeBuffer = &m_buffers[0];
-	m_completeBuffer = m_buffers[0];
-
-	m_threadContext.core->setVideoBuffer(m_threadContext.core, reinterpret_cast<color_t*>(m_activeBuffer->data()), size.width());
-
 	m_resetActions.append([this]() {
 		if (m_autoload) {
 			mCoreLoadState(m_threadContext.core, 0, m_loadStateFlags);
@@ -91,9 +81,11 @@ CoreController::CoreController(mCore* core, QObject* parent)
 
 		controller->m_resetActions.clear();
 
-		controller->m_activeBuffer = &controller->m_buffers[0];
-		context->core->setVideoBuffer(context->core, reinterpret_cast<color_t*>(controller->m_activeBuffer->data()), controller->screenDimensions().width());
+		if (!controller->m_hwaccel) {
+			context->core->setVideoBuffer(context->core, reinterpret_cast<color_t*>(controller->m_activeBuffer.data()), controller->screenDimensions().width());
+		}
 
+		QMetaObject::invokeMethod(controller, "didReset");
 		controller->finishFrame();
 	};
 
@@ -188,7 +180,6 @@ CoreController::CoreController(mCore* core, QObject* parent)
 		message = QString().vsprintf(format, args);
 		QMetaObject::invokeMethod(controller, "logPosted", Q_ARG(int, level), Q_ARG(int, category), Q_ARG(const QString&, message));
 		if (level == mLOG_FATAL) {
-			mCoreThreadMarkCrashed(controller->thread());
 			QMetaObject::invokeMethod(controller, "crashed", Q_ARG(const QString&, QString().vsprintf(format, args)));
 		}
 	};
@@ -211,10 +202,34 @@ CoreController::~CoreController() {
 }
 
 const color_t* CoreController::drawContext() {
-	QMutexLocker locker(&m_mutex);
+	if (m_hwaccel) {
+		return nullptr;
+	}
+	QMutexLocker locker(&m_bufferMutex);
 	return reinterpret_cast<const color_t*>(m_completeBuffer.constData());
 }
 
+QImage CoreController::getPixels() {
+	QByteArray buffer;
+	QSize size = screenDimensions();
+	size_t stride = size.width() * BYTES_PER_PIXEL;
+
+	if (!m_hwaccel) {
+		buffer = m_completeBuffer;
+	} else {
+		Interrupter interrupter(this);
+		const void* pixels;
+		m_threadContext.core->getPixels(m_threadContext.core, &pixels, &stride);
+		stride *= BYTES_PER_PIXEL;
+		buffer = QByteArray::fromRawData(static_cast<const char*>(pixels), stride * size.height());
+	}
+
+	QImage image(reinterpret_cast<const uchar*>(buffer.constData()),
+	             size.width(), size.height(), stride, QImage::Format_RGBX8888);
+	image.bits(); // Cause QImage to detach
+	return image;
+}
+
 bool CoreController::isPaused() {
 	return mCoreThreadIsPaused(&m_threadContext);
 }
@@ -239,6 +254,7 @@ void CoreController::loadConfig(ConfigController* config) {
 	m_loadStateFlags = config->getOption("loadStateExtdata", m_loadStateFlags).toInt();
 	m_saveStateFlags = config->getOption("saveStateExtdata", m_saveStateFlags).toInt();
 	m_fastForwardRatio = config->getOption("fastForwardRatio", m_fastForwardRatio).toFloat();
+	m_fastForwardHeldRatio = config->getOption("fastForwardHeldRatio", m_fastForwardRatio).toFloat();
 	m_videoSync = config->getOption("videoSync", m_videoSync).toInt();
 	m_audioSync = config->getOption("audioSync", m_audioSync).toInt();
 	m_fpsTarget = config->getOption("fpsTarget").toFloat();
@@ -261,6 +277,7 @@ void CoreController::setDebugger(mDebugger* debugger) {
 	Interrupter interrupter(this);
 	if (debugger) {
 		mDebuggerAttach(debugger, m_threadContext.core);
+		mDebuggerEnter(debugger, DEBUGGER_ENTER_ATTACHED, 0);
 	} else {
 		m_threadContext.core->detachDebugger(m_threadContext.core);
 	}
@@ -341,6 +358,15 @@ void CoreController::setLogger(LogController* logger) {
 }
 
 void CoreController::start() {
+	if (!m_hwaccel) {
+		QSize size(256, 224);
+		m_activeBuffer.resize(size.width() * size.height() * sizeof(color_t));
+		m_activeBuffer.fill(0xFF);
+		m_completeBuffer = m_activeBuffer;
+
+		m_threadContext.core->setVideoBuffer(m_threadContext.core, reinterpret_cast<color_t*>(m_activeBuffer.data()), size.width());
+	}
+
 	if (!m_patched) {
 		mCoreAutoloadPatch(m_threadContext.core);
 	}
@@ -357,7 +383,6 @@ void CoreController::stop() {
 #endif
 	setPaused(false);
 	mCoreThreadEnd(&m_threadContext);
-	emit stopping();
 }
 
 void CoreController::reset() {
@@ -375,8 +400,7 @@ void CoreController::setPaused(bool paused) {
 		return;
 	}
 	if (paused) {
-		QMutexLocker locker(&m_mutex);
-		m_frameActions.append([this]() {
+		addFrameAction([this]() {
 			mCoreThreadPauseFromThread(&m_threadContext);
 		});
 	} else {
@@ -385,13 +409,17 @@ void CoreController::setPaused(bool paused) {
 }
 
 void CoreController::frameAdvance() {
-	QMutexLocker locker(&m_mutex);
-	m_frameActions.append([this]() {
+	addFrameAction([this]() {
 		mCoreThreadPauseFromThread(&m_threadContext);
 	});
 	setPaused(false);
 }
 
+void CoreController::addFrameAction(std::function<void ()> action) {
+	QMutexLocker locker(&m_actionMutex);
+	m_frameActions.append(action);
+}
+
 void CoreController::setSync(bool sync) {
 	if (sync) {
 		m_threadContext.impl->sync.audioWait = m_audioSync;
@@ -434,13 +462,21 @@ void CoreController::rewind(int states) {
 }
 
 void CoreController::setFastForward(bool enable) {
+	if (m_fastForward == enable) {
+		return;
+	}
 	m_fastForward = enable;
 	updateFastForward();
+	emit fastForwardChanged(enable);
 }
 
 void CoreController::forceFastForward(bool enable) {
+	if (m_fastForwardForced == enable) {
+		return;
+	}
 	m_fastForwardForced = enable;
 	updateFastForward();
+	emit fastForwardChanged(enable || m_fastForward);
 }
 
 void CoreController::loadState(int slot) {
@@ -593,13 +629,20 @@ void CoreController::replaceGame(const QString& path) {
 }
 
 void CoreController::yankPak() {
-#ifdef M_CORE_GBA
-	if (platform() != PLATFORM_GBA) {
-		return;
-	}
 	Interrupter interrupter(this);
-	GBAYankROM(static_cast<GBA*>(m_threadContext.core->board));
+
+	switch (platform()) {
+#ifdef M_CORE_GBA
+	case PLATFORM_GBA:
+		GBAYankROM(static_cast<GBA*>(m_threadContext.core->board));
+		break;
 #endif
+#ifdef M_CORE_GB
+	case PLATFORM_GB:
+		GBYankROM(static_cast<GB*>(m_threadContext.core->board));
+		break;
+#endif
+	}
 }
 
 void CoreController::addKey(int key) {
@@ -673,8 +716,8 @@ void CoreController::exportSharkport(const QString& path) {
 #endif
 }
 
-void CoreController::attachPrinter() {
 #ifdef M_CORE_GB
+void CoreController::attachPrinter() {
 	if (platform() != PLATFORM_GB) {
 		return;
 	}
@@ -704,11 +747,9 @@ void CoreController::attachPrinter() {
 	};
 	Interrupter interrupter(this);
 	GBSIOSetDriver(&gb->sio, &m_printer.d.d);
-#endif
 }
 
 void CoreController::detachPrinter() {
-#ifdef M_CORE_GB
 	if (platform() != PLATFORM_GB) {
 		return;
 	}
@@ -716,18 +757,52 @@ void CoreController::detachPrinter() {
 	GB* gb = static_cast<GB*>(m_threadContext.core->board);
 	GBPrinterDonePrinting(&m_printer.d);
 	GBSIOSetDriver(&gb->sio, nullptr);
-#endif
 }
 
 void CoreController::endPrint() {
-#ifdef M_CORE_GB
 	if (platform() != PLATFORM_GB) {
 		return;
 	}
 	Interrupter interrupter(this);
 	GBPrinterDonePrinting(&m_printer.d);
-#endif
 }
+#endif
+
+#ifdef M_CORE_GBA
+void CoreController::attachBattleChipGate() {
+	if (platform() != PLATFORM_GBA) {
+		return;
+	}
+	Interrupter interrupter(this);
+	clearMultiplayerController();
+	GBASIOBattlechipGateCreate(&m_battlechip);
+	m_threadContext.core->setPeripheral(m_threadContext.core, mPERIPH_GBA_BATTLECHIP_GATE, &m_battlechip);
+}
+
+void CoreController::detachBattleChipGate() {
+	if (platform() != PLATFORM_GBA) {
+		return;
+	}
+	Interrupter interrupter(this);
+	m_threadContext.core->setPeripheral(m_threadContext.core, mPERIPH_GBA_BATTLECHIP_GATE, nullptr);
+}
+
+void CoreController::setBattleChipId(uint16_t id) {
+	if (platform() != PLATFORM_GBA) {
+		return;
+	}
+	Interrupter interrupter(this);
+	m_battlechip.chipId = id;
+}
+
+void CoreController::setBattleChipFlavor(int flavor) {
+	if (platform() != PLATFORM_GBA) {
+		return;
+	}
+	Interrupter interrupter(this);
+	m_battlechip.flavor = flavor;
+}
+#endif
 
 void CoreController::setAVStream(mAVStream* stream) {
 	Interrupter interrupter(this);
@@ -743,32 +818,55 @@ void CoreController::clearOverride() {
 	m_override.reset();
 }
 
-void CoreController::startVideoLog(const QString& path) {
+void CoreController::startVideoLog(const QString& path, bool compression) {
 	if (m_vl) {
 		return;
 	}
 
+	VFile* vf = VFileDevice::open(path, O_WRONLY | O_CREAT | O_TRUNC);
+	if (!vf) {
+		return;
+	}
+	startVideoLog(vf);
+}
+
+void CoreController::startVideoLog(VFile* vf, bool compression) {
+	if (m_vl || !vf) {
+		return;
+	}
+
 	Interrupter interrupter(this);
 	m_vl = mVideoLogContextCreate(m_threadContext.core);
-	m_vlVf = VFileDevice::open(path, O_WRONLY | O_CREAT | O_TRUNC);
+	m_vlVf = vf;
 	mVideoLogContextSetOutput(m_vl, m_vlVf);
+	mVideoLogContextSetCompression(m_vl, compression);
 	mVideoLogContextWriteHeader(m_vl, m_threadContext.core);
 }
 
-void CoreController::endVideoLog() {
+void CoreController::endVideoLog(bool closeVf) {
 	if (!m_vl) {
 		return;
 	}
 
 	Interrupter interrupter(this);
 	mVideoLogContextDestroy(m_threadContext.core, m_vl);
-	if (m_vlVf) {
+	if (m_vlVf && closeVf) {
 		m_vlVf->close(m_vlVf);
 		m_vlVf = nullptr;
 	}
 	m_vl = nullptr;
 }
 
+void CoreController::setFramebufferHandle(int fb) {
+	Interrupter interrupter(this);
+	if (fb < 0) {
+		m_hwaccel = false;
+	} else {
+		m_threadContext.core->setVideoGLTex(m_threadContext.core, fb);
+		m_hwaccel = true;
+	}
+}
+
 void CoreController::updateKeys() {
 	int activeKeys = m_activeKeys | updateAutofire() | m_inputController->pollEvents();
 	m_threadContext.core->setKeys(m_threadContext.core, activeKeys);
@@ -791,28 +889,28 @@ int CoreController::updateAutofire() {
 }
 
 void CoreController::finishFrame() {
-	QMutexLocker locker(&m_mutex);
-	memcpy(m_completeBuffer.data(), m_activeBuffer->constData(), m_activeBuffer->size());
+	if (!m_hwaccel) {
+		unsigned width, height;
+		m_threadContext.core->desiredVideoDimensions(m_threadContext.core, &width, &height);
 
-	// TODO: Generalize this to triple buffering?
-	m_activeBuffer = &m_buffers[0];
-	if (m_activeBuffer == m_completeBuffer) {
-		m_activeBuffer = &m_buffers[1];
+		QMutexLocker locker(&m_bufferMutex);
+		memcpy(m_completeBuffer.data(), m_activeBuffer.constData(), 256 * height * BYTES_PER_PIXEL);
 	}
-	// Copy contents to avoid issues when doing frameskip
-	memcpy(m_activeBuffer->data(), m_completeBuffer.constData(), m_activeBuffer->size());
-	m_threadContext.core->setVideoBuffer(m_threadContext.core, reinterpret_cast<color_t*>(m_activeBuffer->data()), screenDimensions().width());
 
-	for (auto& action : m_frameActions) {
+	QMutexLocker locker(&m_actionMutex);
+	QList<std::function<void ()>> frameActions(m_frameActions);
+	m_frameActions.clear();
+	for (auto& action : frameActions) {
 		action();
 	}
-	m_frameActions.clear();
 	updateKeys();
 
 	QMetaObject::invokeMethod(this, "frameAvailable");
 }
 
 void CoreController::updateFastForward() {
+	// If we have "Fast forward" checked in the menu (m_fastForwardForced)
+	// or are holding the fast forward button (m_fastForward):
 	if (m_fastForward || m_fastForwardForced) {
 		if (m_fastForwardVolume >= 0) {
 			m_threadContext.core->opts.volume = m_fastForwardVolume;
@@ -820,10 +918,25 @@ void CoreController::updateFastForward() {
 		if (m_fastForwardMute >= 0) {
 			m_threadContext.core->opts.mute = m_fastForwardMute;
 		}
-		if (m_fastForwardRatio > 0) {
-			m_threadContext.impl->sync.fpsTarget = m_fpsTarget * m_fastForwardRatio;
+
+		// If we aren't holding the fast forward button
+		// then use the non "(held)" ratio
+		if(!m_fastForward) {
+			if (m_fastForwardRatio > 0) {
+				m_threadContext.impl->sync.fpsTarget = m_fpsTarget * m_fastForwardRatio;
+				setSync(true);
+			}	else {
+				setSync(false);
+			}
 		} else {
-			setSync(false);
+			// If we are holding the fast forward button,
+			// then use the held ratio
+			if (m_fastForwardHeldRatio > 0) {
+				m_threadContext.impl->sync.fpsTarget = m_fpsTarget * m_fastForwardHeldRatio;
+				setSync(true);
+			} else {
+				setSync(false);
+			}
 		}
 	} else {
 		if (!mCoreConfigGetIntValue(&m_threadContext.core->config, "volume", &m_threadContext.core->opts.volume)) {
@@ -835,8 +948,8 @@ void CoreController::updateFastForward() {
 		m_threadContext.impl->sync.fpsTarget = m_fpsTarget;
 		setSync(true);
 	}
-	// XXX: Have a way of just updating opts
-	m_threadContext.core->loadConfig(m_threadContext.core, &m_threadContext.core->config);
+
+	m_threadContext.core->reloadConfigOption(m_threadContext.core, NULL, NULL);
 }
 
 CoreController::Interrupter::Interrupter(CoreController* parent, bool fromThread)
