@@ -88,6 +88,19 @@ MemoryModel::MemoryModel(QWidget* parent)
 		update();
 	});
 
+	connect(verticalScrollBar(), &QSlider::actionTriggered, [this](int action) {
+		if (action == QSlider::SliderSingleStepAdd) {
+			++m_top;
+		} else if (action == QSlider::SliderSingleStepSub) {
+			--m_top;
+		} else {
+			return;
+		}
+		boundsCheck();
+		verticalScrollBar()->setValue(m_top);
+		update();
+	});
+
 	setRegion(0, 0x10000000, tr("All"));
 }
 
