@@ -17,6 +17,7 @@
 #include <mgba/core/blip_buf.h>
 #include <mgba/core/core.h>
 #include "feature/gui/gui-runner.h"
+#include <mgba/internal/gb/video.h>
 #include <mgba/internal/gba/audio.h>
 #include <mgba/internal/gba/gba.h>
 #include <mgba/internal/gba/input.h>
@@ -63,7 +64,8 @@ static enum VideoMode {
 	VM_MAX
 } videoMode = VM_AUTODETECT;
 
-#define SAMPLES 1024
+#define SAMPLES 512
+#define BUFFERS 8
 #define GUI_SCALE 1.35f
 #define GUI_SCALE_240p 2.0f
 
@@ -87,6 +89,7 @@ static void _setup(struct mGUIRunner* runner);
 static void _gameLoaded(struct mGUIRunner* runner);
 static void _gameUnloaded(struct mGUIRunner* runner);
 static void _unpaused(struct mGUIRunner* runner);
+static void _prepareForFrame(struct mGUIRunner* runner);
 static void _drawFrame(struct mGUIRunner* runner, bool faded);
 static uint16_t _pollGameInput(struct mGUIRunner* runner);
 static void _setFrameLimiter(struct mGUIRunner* runner, bool limit);
@@ -102,7 +105,7 @@ static struct mRotationSource rotation;
 static GXRModeObj* vmode;
 static float wAdjust;
 static float hAdjust;
-static float wStretch = 1.0f;
+static float wStretch = 0.9f;
 static float hStretch = 0.9f;
 static float guiScale = GUI_SCALE;
 static Mtx model, view, modelview;
@@ -110,6 +113,9 @@ static uint16_t* texmem;
 static GXTexObj tex;
 static uint16_t* rescaleTexmem;
 static GXTexObj rescaleTex;
+static uint16_t* interframeTexmem;
+static GXTexObj interframeTex;
+static bool sgbCrop = false;
 static int32_t tiltX;
 static int32_t tiltY;
 static int32_t gyroZ;
@@ -118,6 +124,7 @@ static uint32_t referenceRetraceCount;
 static bool frameLimiter = true;
 static int scaleFactor;
 static unsigned corew, coreh;
+static bool interframeBlending = true;
 
 uint32_t* romBuffer;
 size_t romBufferSize;
@@ -125,9 +132,12 @@ size_t romBufferSize;
 static void* framebuffer[2] = { 0, 0 };
 static int whichFb = 0;
 
-static struct GBAStereoSample audioBuffer[3][SAMPLES] __attribute__((__aligned__(32)));
-static volatile size_t audioBufferSize = 0;
+static struct AudioBuffer {
+	struct GBAStereoSample samples[SAMPLES] __attribute__((__aligned__(32)));
+	volatile size_t size;
+} audioBuffer[BUFFERS] = {0};
 static volatile int currentAudioBuffer = 0;
+static volatile int nextAudioBuffer = 0;
 static double audioSampleRate = 60.0 / 1.001;
 
 static struct GUIFont* font;
@@ -195,6 +205,9 @@ static void reconfigureScreen(struct mGUIRunner* runner) {
 		break;
 	}
 
+	vmode->viWidth = 704;
+	vmode->viXOrigin = 8;
+
 	VIDEO_SetBlack(true);
 	VIDEO_Configure(vmode);
 
@@ -274,10 +287,16 @@ int main(int argc, char* argv[]) {
 	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_S16, 0);
 	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
 
+	GX_SetNumTevStages(1);
 	GX_SetNumChans(1);
 	GX_SetNumTexGens(1);
 	GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
+	GX_SetTevOrder(GX_TEVSTAGE1, GX_TEXCOORD0, GX_TEXMAP1, GX_COLOR0A0);
 	GX_SetTevOp(GX_TEVSTAGE0, GX_MODULATE);
+	GX_SetTevColorOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_DIVIDE_2, GX_TRUE, GX_TEVPREV);
+	GX_SetTevAlphaOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV);
+	GX_SetTevColorIn(GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_TEXC, GX_CC_ONE, GX_CC_CPREV);
+	GX_SetTevAlphaIn(GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_APREV);
 
 	GX_SetTexCoordGen(GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);
 	GX_InvVtxCache();
@@ -295,6 +314,8 @@ int main(int argc, char* argv[]) {
 
 	texmem = memalign(32, TEX_W * TEX_H * BYTES_PER_PIXEL);
 	GX_InitTexObj(&tex, texmem, TEX_W, TEX_H, GX_TF_RGB565, GX_CLAMP, GX_CLAMP, GX_FALSE);
+	interframeTexmem = memalign(32, TEX_W * TEX_H * BYTES_PER_PIXEL);
+	GX_InitTexObj(&interframeTex, interframeTexmem, TEX_W, TEX_H, GX_TF_RGB565, GX_CLAMP, GX_CLAMP, GX_FALSE);
 	rescaleTexmem = memalign(32, TEX_W * TEX_H * 4 * BYTES_PER_PIXEL);
 	GX_InitTexObj(&rescaleTex, rescaleTexmem, TEX_W * 2, TEX_H * 2, GX_TF_RGB565, GX_CLAMP, GX_CLAMP, GX_FALSE);
 	GX_InitTexObjFilterMode(&rescaleTex, GX_LINEAR, GX_LINEAR);
@@ -319,7 +340,7 @@ int main(int argc, char* argv[]) {
 
 	struct mGUIRunner runner = {
 		.params = {
-			720, 480,
+			640, 480,
 			font, "",
 			_drawStart, _drawEnd,
 			_pollInput, _pollCursor,
@@ -476,12 +497,12 @@ int main(int argc, char* argv[]) {
 				.submenu = 0,
 				.state = 7,
 				.validStates = (const char*[]) {
-					"1/2x", "0.6x", "1/3x", "0.7x", "1/4x", "0.8x", "0.9x", "1.0x"
+					"1/2x", "0.6x", "2/3x", "0.7x", "3/4x", "0.8x", "0.9x", "1.0x"
 				},
 				.stateMappings = (const struct GUIVariant[]) {
 					GUI_V_F(0.5f),
 					GUI_V_F(0.6f),
-					GUI_V_F(1.f / 3.f),
+					GUI_V_F(2.f / 3.f),
 					GUI_V_F(0.7f),
 					GUI_V_F(0.75f),
 					GUI_V_F(0.8f),
@@ -496,12 +517,12 @@ int main(int argc, char* argv[]) {
 				.submenu = 0,
 				.state = 6,
 				.validStates = (const char*[]) {
-					"1/2x", "0.6x", "1/3x", "0.7x", "1/4x", "0.8x", "0.9x", "1.0x"
+					"1/2x", "0.6x", "2/3x", "0.7x", "3/4x", "0.8x", "0.9x", "1.0x"
 				},
 				.stateMappings = (const struct GUIVariant[]) {
 					GUI_V_F(0.5f),
 					GUI_V_F(0.6f),
-					GUI_V_F(1.f / 3.f),
+					GUI_V_F(2.f / 3.f),
 					GUI_V_F(0.7f),
 					GUI_V_F(0.75f),
 					GUI_V_F(0.8f),
@@ -516,7 +537,7 @@ int main(int argc, char* argv[]) {
 		.teardown = 0,
 		.gameLoaded = _gameLoaded,
 		.gameUnloaded = _gameUnloaded,
-		.prepareForFrame = 0,
+		.prepareForFrame = _prepareForFrame,
 		.drawFrame = _drawFrame,
 		.paused = _gameUnloaded,
 		.unpaused = _unpaused,
@@ -548,9 +569,7 @@ int main(int argc, char* argv[]) {
 	_mapKey(&runner.params.keyMap, WIIMOTE_INPUT, WPAD_BUTTON_DOWN, GUI_INPUT_RIGHT);
 
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_A, GUI_INPUT_SELECT);
-	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_Y, GUI_INPUT_SELECT);
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_B, GUI_INPUT_BACK);
-	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_X, GUI_INPUT_BACK);
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_HOME, GUI_INPUT_CANCEL);
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_UP, GUI_INPUT_UP);
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_DOWN, GUI_INPUT_DOWN);
@@ -583,6 +602,7 @@ int main(int argc, char* argv[]) {
 	free(fifo);
 	free(texmem);
 	free(rescaleTexmem);
+	free(interframeTexmem);
 
 	free(outputBuffer);
 	GUIFontDestroy(font);
@@ -594,32 +614,43 @@ int main(int argc, char* argv[]) {
 }
 
 static void _audioDMA(void) {
-	if (!audioBufferSize) {
+	struct AudioBuffer* buffer = &audioBuffer[currentAudioBuffer];
+	if (buffer->size != SAMPLES) {
 		return;
 	}
-	DCFlushRange(audioBuffer[currentAudioBuffer], audioBufferSize * sizeof(struct GBAStereoSample));
-	AUDIO_InitDMA((u32) audioBuffer[currentAudioBuffer], audioBufferSize * sizeof(struct GBAStereoSample));
-	currentAudioBuffer = (currentAudioBuffer + 1) % 3;
-	audioBufferSize = 0;
+	DCFlushRange(buffer->samples, SAMPLES * sizeof(struct GBAStereoSample));
+	AUDIO_InitDMA((u32) buffer->samples, SAMPLES * sizeof(struct GBAStereoSample));
+	buffer->size = 0;
+	currentAudioBuffer = (currentAudioBuffer + 1) % BUFFERS;
 }
 
 static void _postAudioBuffer(struct mAVStream* stream, blip_t* left, blip_t* right) {
 	UNUSED(stream);
+
+	u32 level = 0;
+	_CPU_ISR_Disable(level);
+	struct AudioBuffer* buffer = &audioBuffer[nextAudioBuffer];
 	int available = blip_samples_avail(left);
-	if (available + audioBufferSize > SAMPLES) {
-		available = SAMPLES - audioBufferSize;
+	if (available + buffer->size > SAMPLES) {
+		available = SAMPLES - buffer->size;
 	}
-	available &= ~((32 / sizeof(struct GBAStereoSample)) - 1); // Force align to 32 bytes
 	if (available > 0) {
 		// These appear to be reversed for AUDIO_InitDMA
-		blip_read_samples(left, &audioBuffer[currentAudioBuffer][audioBufferSize].right, available, true);
-		blip_read_samples(right, &audioBuffer[currentAudioBuffer][audioBufferSize].left, available, true);
-		audioBufferSize += available;
+		blip_read_samples(left, &buffer->samples[buffer->size].right, available, true);
+		blip_read_samples(right, &buffer->samples[buffer->size].left, available, true);
+		buffer->size += available;
 	}
-	if (audioBufferSize == SAMPLES && !AUDIO_GetDMAEnableFlag()) {
-		_audioDMA();
-		AUDIO_StartDMA();
+	if (buffer->size == SAMPLES) {
+		int next = (nextAudioBuffer + 1) % BUFFERS;
+		if ((currentAudioBuffer + BUFFERS - next) % BUFFERS != 1) {
+			nextAudioBuffer = next;
+		}
+		if (!AUDIO_GetDMAEnableFlag()) {
+			_audioDMA();
+			AUDIO_StartDMA();
+		}
 	}
+	_CPU_ISR_Restore(level);
 }
 
 static void _drawStart(void) {
@@ -632,6 +663,8 @@ static void _drawStart(void) {
 			VIDEO_WaitVSync();
 		}
 		referenceRetraceCount = retraceCount;
+	} else if (frameLimiter && referenceRetraceCount < retraceCount - 1) {
+		referenceRetraceCount = retraceCount - 1;
 	}
 	_CPU_ISR_Restore(level);
 
@@ -727,6 +760,7 @@ void _reproj2(int w, int h) {
 }
 
 void _guiPrepare(void) {
+	GX_SetNumTevStages(1);
 	_reproj2(vmode->fbWidth * guiScale * wAdjust, vmode->efbHeight * guiScale * hAdjust);
 }
 
@@ -779,6 +813,12 @@ void _setup(struct mGUIRunner* runner) {
 	outputBuffer = memalign(32, TEX_W * TEX_H * BYTES_PER_PIXEL);
 	runner->core->setVideoBuffer(runner->core, outputBuffer, TEX_W);
 
+	nextAudioBuffer = 0;
+	currentAudioBuffer = 0;
+	int i;
+	for (i = 0; i < BUFFERS; ++i) {
+		audioBuffer[i].size = 0;
+	}
 	runner->core->setAudioBufferSize(runner->core, SAMPLES);
 
 	double ratio = GBAAudioCalculateRatio(1, audioSampleRate, 1);
@@ -810,6 +850,7 @@ void _gameLoaded(struct mGUIRunner* runner) {
 		}
 	}
 	memset(texmem, 0, TEX_W * TEX_H * BYTES_PER_PIXEL);
+	memset(interframeTexmem, 0, TEX_W * TEX_H * BYTES_PER_PIXEL);
 	_unpaused(runner);
 }
 
@@ -842,6 +883,14 @@ void _unpaused(struct mGUIRunner* runner) {
 			break;
 		}
 	}
+	int fakeBool;
+	if (mCoreConfigGetIntValue(&runner->config, "interframeBlending", &fakeBool)) {
+		interframeBlending = fakeBool;
+	}
+	if (mCoreConfigGetIntValue(&runner->config, "sgb.borderCrop", &fakeBool)) {
+		sgbCrop = fakeBool;
+	}
+
 	float stretch;
 	if (mCoreConfigGetFloatValue(&runner->config, "stretchWidth", &stretch)) {
 		wStretch = fminf(1.0f, fmaxf(0.5f, stretch));
@@ -851,6 +900,12 @@ void _unpaused(struct mGUIRunner* runner) {
 	}
 }
 
+void _prepareForFrame(struct mGUIRunner* runner) {
+	if (interframeBlending) {
+		memcpy(interframeTexmem, texmem, TEX_W * TEX_H * BYTES_PER_PIXEL);
+	}
+}
+
 void _drawFrame(struct mGUIRunner* runner, bool faded) {
 	runner->core->desiredVideoDimensions(runner->core, &corew, &coreh);
 	uint32_t color = 0xFFFFFF3F;
@@ -869,14 +924,24 @@ void _drawFrame(struct mGUIRunner* runner, bool faded) {
 		}
 	}
 	DCFlushRange(texdest, TEX_W * TEX_H * BYTES_PER_PIXEL);
+	if (interframeBlending) {
+		DCFlushRange(interframeTexmem, TEX_W * TEX_H * BYTES_PER_PIXEL);
+	}
 
-	if (faded) {
+	if (faded || interframeBlending) {
 		GX_SetBlendMode(GX_BM_BLEND, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_NOOP);
 	} else {
 		GX_SetBlendMode(GX_BM_NONE, GX_BL_ONE, GX_BL_ZERO, GX_LO_NOOP);
 	}
 	GX_InvalidateTexAll();
-	GX_LoadTexObj(&tex, GX_TEXMAP0);
+	if (interframeBlending) {
+		GX_LoadTexObj(&interframeTex, GX_TEXMAP0);
+		GX_LoadTexObj(&tex, GX_TEXMAP1);
+		GX_SetNumTevStages(2);
+	} else {
+		GX_LoadTexObj(&tex, GX_TEXMAP0);
+		GX_SetNumTevStages(1);
+	}
 
 	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
 	s16 vertWidth = corew;
@@ -909,22 +974,31 @@ void _drawFrame(struct mGUIRunner* runner, bool faded) {
 		GX_SetTexCopyDst(TEX_W * 2, TEX_H * 2, GX_TF_RGB565, GX_FALSE);
 		GX_CopyTex(rescaleTexmem, GX_TRUE);
 		GX_LoadTexObj(&rescaleTex, GX_TEXMAP0);
-	}
-
-	int hfactor = (vmode->fbWidth * wStretch) / (corew * wAdjust);
-	int vfactor = (vmode->efbHeight * hStretch) / (coreh * hAdjust);
-	if (hfactor > vfactor) {
-		scaleFactor = vfactor;
-	} else {
-		scaleFactor = hfactor;
+		GX_SetNumTevStages(1);
+		if (!faded) {
+			GX_SetBlendMode(GX_BM_NONE, GX_BL_ONE, GX_BL_ZERO, GX_LO_NOOP);
+		}
 	}
 
 	if (screenMode == SM_PA) {
+		unsigned factorWidth = corew;
+		unsigned factorHeight = coreh;
+		if (sgbCrop && factorWidth == 256 && factorHeight == 224) {
+			factorWidth = GB_VIDEO_HORIZONTAL_PIXELS;
+			factorHeight = GB_VIDEO_VERTICAL_PIXELS;
+		}
+
+		int hfactor = (vmode->fbWidth * wStretch) / (factorWidth * wAdjust);
+		int vfactor = (vmode->efbHeight * hStretch) / (factorHeight * hAdjust);
+		if (hfactor > vfactor) {
+			scaleFactor = vfactor;
+		} else {
+			scaleFactor = hfactor;
+		}
+
 		vertWidth *= scaleFactor;
 		vertHeight *= scaleFactor;
-	}
 
-	if (screenMode == SM_PA) {
 		_reproj(corew * scaleFactor, coreh * scaleFactor);
 	} else {
 		_reproj2(corew, coreh);
@@ -1074,8 +1148,15 @@ static s8 WPAD_StickX(u8 chan, u8 right) {
 		return 0;
 	}
 	int centered = (int) js->pos.x - (int) js->center.x;
-	int range = js->max.x - js->min.x;
-	return (centered * 0xFF) / range;
+	int range = (int) js->max.x - (int) js->min.x;
+	int value = (centered * 0xFF) / range;
+	if (value > 0x7F) {
+		return 0x7F;
+	}
+	if (value < -0x80) {
+		return -0x80;
+	}
+	return value;
 }
 
 static s8 WPAD_StickY(u8 chan, u8 right) {
@@ -1105,8 +1186,15 @@ static s8 WPAD_StickY(u8 chan, u8 right) {
 		return 0;
 	}
 	int centered = (int) js->pos.y - (int) js->center.y;
-	int range = js->max.y - js->min.y;
-	return (centered * 0xFF) / range;
+	int range = (int) js->max.y - (int) js->min.y;
+	int value = (centered * 0xFF) / range;
+	if (value > 0x7F) {
+		return 0x7F;
+	}
+	if (value < -0x80) {
+		return -0x80;
+	}
+	return value;
 }
 
 void _retraceCallback(u32 count) {
